#!/usr/bin/env python

'''
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import threading
from math import sin, pi

from pyglet.gl import *
from pyglet import clock
from pyglet import event
from pyglet import image
from pyglet import window
from pyglet.window import key

import euclid

import susan
import view
import menu
from res import getres

class GameWindow(window.Window):
    player_indicator_materials = [
        view.Material((0, 0, .6, 1),
                      (0, 0, .6, .5),
                      (1, 1, 1, 1),
                      3,),
        view.Material((.6, 0, 0, 1),
                      (.6, 0, 0, .5),
                      (1, 1, 1, 1),
                      3,),
    ]

    def __init__(self, *args, **kwargs):
        super(GameWindow, self).__init__(*args, **kwargs)

        glEnable(GL_DEPTH_TEST)
        glEnable(GL_CULL_FACE)
        self.control_stack = []
        self.current_player = 0

        self.menu_view = view.BoardView(menu.MenuLayout(), getres('menu.png'))
        self.board_view = view.BoardView(susan.BoardLayout([4, 2, 2]),
                                         getres('board3.png'))

        self.camera_tx = 0.
        self.camera_ty = 0.
        self.camera_rx = -40.
        self.camera_ry = 0.
        self.reset_indicator()
        self.draw_indicator = False

    def reset_indicator(self):
        self.indicator_ry = 0.
        self.indicator_player = 0
        self.indicator_stop = True
        self.indicator_thinking = False
        r = 32
        self.indicator_x = self.width - r
        self.indicator_y = self.height - r
        self.indicator_z = -r * 2
        self.indicator_scale = r * 2

    def set_2d(self, near, far):
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        glOrtho(0, self.width, 0, self.height, near, far)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()

    def unset_2d(self):
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW) 

    def on_resize(self, width, height):
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(40., self.width / float(self.height), 1, 100.)
        glMatrixMode(GL_MODELVIEW)
        self.reset_indicator()

    def begin_intro(self):
        self.view = self.menu_view
        self.layout = self.view.layout
        self.menu_state = self.layout.initial()
        self.setup_menu()
        self.begin_menu()
        self.push_control(IntroControl(self))

    def setup_game(self):
        self.view = self.board_view
        self.layout = self.view.layout
        self.state = self.layout.initial()
        self.current_player = 0
        self.view.reset_angles()

    def setup_menu(self):
        self.view = self.menu_view
        self.layout = self.view.layout
        self.state = self.menu_state

    def begin_game(self):
        def delay(dt):
            self.draw_indicator = True
            self.push_control(self.players[self.current_player])
            for player in self.players:
                player.begin_game()
            self.players[self.current_player].begin_turn()
        clock.schedule_once(delay, 1.)

    def begin_menu(self):
        self.push_control(MenuControl(self))
        self.draw_indicator = False

    def next_player(self):
        self.pop_control()

        self.indicator_thinking = False
        self.indicator_stop = False

        if self.layout.is_won(self.state):
            self.push_control(GameOverDialog(self))
            return

        self.current_player = 1 - self.current_player
        self.push_control(self.players[self.current_player])
        self.players[self.current_player].begin_turn()

    def push_control(self, control):
        self.control_stack.append(control)
        self.push_handlers(control)

    def pop_control(self):
        self.control_stack.pop()
        self.pop_handlers()
        if self.control_stack:
            self.control_stack[-1].regain_control()

    def update(self, dt):
        for controller in self.control_stack[::-1]:
            if controller.update(dt):
                return
        if (self.indicator_thinking or 
            self.indicator_ry < 0 or 
            self.indicator_player != self.current_player):
            self.indicator_ry += dt * 180 
            if self.indicator_ry > 90:
                self.indicator_ry -= 180.
                self.indicator_player = self.current_player
                self.indicator_stop = not self.indicator_thinking
            elif self.indicator_ry > 0 and self.indicator_stop:
                self.indicator_ry = 0

    def draw(self):
        glClearColor(.7, .7, .7, 1)
        self.clear()

        glLoadIdentity()
        glTranslatef(self.camera_tx, self.camera_ty, -8.)
        glRotatef(self.camera_rx, 1, 0, 0)
        glRotatef(self.camera_ry, 0, 1, 0)
        glTranslatef(0, 0, view.RING_HEIGHT / 2)

        self.view.draw(self.state)

        if self.draw_indicator:
            r = 32
            glPushAttrib(GL_ENABLE_BIT)
            self.set_2d(0, r*4)
            glDisable(GL_DEPTH_TEST)
            glTranslatef(self.indicator_x, self.indicator_y, self.indicator_z)
            glRotatef(self.indicator_ry, 0, 1, 0)
            glScalef(self.indicator_scale,
                     self.indicator_scale,
                     self.indicator_scale)
            glTranslatef(0, 0, -view.PIECE_HEIGHT/2)
            self.player_indicator_materials[self.indicator_player].apply()
            glCallList(self.view.piece_list)
            self.unset_2d()
            glPopAttrib()

        for controller in self.control_stack:
            controller.draw()

        self.flip()

    def apply(self, turn):
        if isinstance(turn, susan.TwistRing):
            self.push_control(AnimateTwistRingTurn(self, turn))
        elif isinstance(turn, susan.MovePiece):
            self.push_control(AnimateMovePieceTurn(self, turn))
        else:
            self.apply_immediate(turn)

    def apply_immediate(self, turn):
        self.state = turn.apply(self.layout, self.state)
        self.control_stack[-1].after_apply()

    def select_ray(self, x, y):
        model_view = (GLdouble * 16)()
        glGetDoublev(GL_MODELVIEW_MATRIX, model_view)
        projection = (GLdouble * 16)()
        glGetDoublev(GL_PROJECTION_MATRIX, projection)
        viewport = (GLint * 4)()
        glGetIntegerv(GL_VIEWPORT, viewport)

        x1, y1, z1 = GLdouble(), GLdouble(), GLdouble()
        x2, y2, z2 = GLdouble(), GLdouble(), GLdouble()
        gluUnProject(x, y, 0, model_view, projection, viewport, x1, y1, z1)
        gluUnProject(x, y, 1, model_view, projection, viewport, x2, y2, z2)
        ray = euclid.Ray3(euclid.Point3(x1.value, y1.value, z1.value), 
                          euclid.Point3(x2.value, y2.value, z2.value))
        ray.v.normalize()
        return ray

    def dispatch_event(self, event_type, *args):
        '''Corrected method from pyglet r1230, fixes infinite recursion
        when event stack is modified during event dispatch.'''
        assert event_type in self.event_types

        # Search handler stack for matching event handlers
        for frame in self._event_stack[:]:
            handler = frame.get(event_type, None)
            if handler:
                ret = handler(*args)
                if ret != event.EVENT_UNHANDLED:
                    return

        # Check instance for an event handler
        if hasattr(self, event_type):
            getattr(self, event_type)(*args)

    def do_menu(self):
        self.push_control(FlipBoardToMenuControl(self))

    def do_game(self):
        self.push_control(FlipBoardToGameControl(self))

    def do_hint(self, *textures):
        self.push_control(HintControl(self, *textures))

class Controller(object):
    def update(self, dt):
        pass

    def draw(self):
        pass

    def after_apply(self):
        pass

    def regain_control(self):
        pass

def interp(a, b, t):
    f = sin((t - .5) * pi) * .5 + .5
    return a + f * (b - a)

def interp1(a, b, t):
    f = sin((t - 1) * pi/2) + 1
    return a + f * (b - a)

def interp2(a, b, t):
    f = sin(t * pi/2)
    return a + f * (b - a)

class BlockInputControl(Controller):
    def on_mouse_press(self, x, y, button, modifiers):
        return True

    def on_mouse_release(self, x, y, button, modifiers):
        return True

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        return True

class HintControl(BlockInputControl):
    def __init__(self, win, *textures):
        self.window = win
        self.textures = [image.load(texture).texture for texture in textures]
        self.texture = None
        self.fade = 0.

    def update(self, dt):
        if not self.texture:
            if self.textures:
                self.fade += dt
                if self.fade >= 1.:
                    self.fade = 1.
                    self.texture = self.textures.pop(0)
            else:
                self.fade -= dt
                if self.fade <= 0:
                    self.fade = 0.
                    self.window.pop_control()

    def draw(self):
        self.window.set_2d(-1, 1)
        glPushAttrib(GL_ENABLE_BIT)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glEnable(GL_BLEND)
        glColor4f(0, 0, 0, self.fade * .5)
        glBegin(GL_QUADS)
        glVertex2f(0, 0)
        glVertex2f(self.window.width, 0)
        glVertex2f(self.window.width, self.window.height)
        glVertex2f(0, self.window.height)
        glEnd()
        glColor4f(1, 1, 1, 1)
        if self.texture:
            self.texture.blit((self.window.width - self.texture.width) // 2,
                              (self.window.height - self.texture.height) // 2)
        glPopAttrib()
        self.window.unset_2d()

    def on_mouse_press(self, x, y, button, modifiers):
        if self.texture:
            if self.textures:
                self.texture = self.textures.pop(0)
            else:
                self.texture = None
        return True

class FlipBoardControl(BlockInputControl):
    vel = 180.

    def __init__(self, win):
        self.window = win
        self.t = 0.

class FlipBoardToGameControl(FlipBoardControl):
    def update(self, dt):
        self.t += dt
        self.window.camera_ry = interp1(0, 91, self.t)
        if self.window.camera_ry > 90.:
            self.window.setup_game()
            self.window.camera_ry -= 180.
            self.window.pop_control()
            self.window.push_control(
                FlipBoardCompleteControl(self.window, self.window.begin_game))

class FlipBoardToMenuControl(FlipBoardControl):
    def update(self, dt):
        self.t += dt
        self.window.camera_ry = interp1(0, 91, self.t)
        if self.window.camera_ry > 90.:
            self.window.setup_menu()
            self.window.camera_ry -= 180.
            self.window.pop_control()
            self.window.push_control(
                FlipBoardCompleteControl(self.window, self.window.begin_menu))

class FlipBoardCompleteControl(FlipBoardControl):
    def __init__(self, win, after):
        super(FlipBoardCompleteControl, self).__init__(win)
        self.after_method = after

    def update(self, dt):
        self.t += dt
        self.window.camera_ry = interp2(-90, 1, self.t)
        if self.window.camera_ry > 0.:
            self.window.camera_ry = 0.
            self.window.pop_control()
            self.after_method()

class IntroControl(BlockInputControl):
    def __init__(self, win):
        self.window = win
        self.t = 0.

    def update(self, dt):
        self.t += dt * 1.5
        if self.t >= 12:
            self.t = 12
            self.window.pop_control()
        def interp(start, finish, starttime, endtime):
            if self.t < starttime:
                return start
            if self.t > endtime:
                return finish
            f = (self.t - starttime) / (endtime - starttime)
            f = sin((f - .5) * pi) * .5 + .5
            return start +  f * (finish - start)
        self.window.camera_rx = interp(0, -40, 3, 10)
        self.window.camera_ry = interp(-40, 0, 3, 12)
        self.window.camera_tx = interp(7, 0, 0, 6)
        self.window.camera_ty = interp(-5, .5, 4, 10)
        self.window.view.ring_angles[2] = interp(-80, 0, 0, 12)
        self.window.view.ring_angles[1] = interp(120, 0, 0, 12)
        self.window.view.ring_angles[0] = interp(180, 0, 8, 12)
         

class MenuControl(Controller):
    first_time = True

    def __init__(self, win):
        self.window = win
        self.inner_dz = 0.
        self.t = 0.

        if self.first_time:
            clock.schedule_once(self.do_hint1, 20)

    def do_hint1(self, dt):
        if not self.first_time:
            return
        self.window.do_hint(getres('menuhint1.png'))

    def do_hint2(self, dt):
        if not self.first_time:
            return
        self.window.do_hint(getres('menuhint2.png'))

    def regain_control(self):
        self.update_inner_disc()

    def update(self, dt):
        self.t += self.inner_dz * dt
        self.t = min(1, max(self.t, 0))
        self.window.view.inner_disc_z = \
            interp(0, view.INNER_DISC_RAISE_Z, self.t)

    def on_mouse_press(self, x, y, button, modifiers):
        ray = self.window.select_ray(x, y)
        cell, pos = self.window.view.get_intersection_piece(ray)
        if cell is not None:
            self.window.view.drag_cell = cell
            self.window.push_control(DragPiece(self.window, cell, pos))
            self.window.dispatch_event('on_mouse_drag', x, y, 0, 0, 0, 0)
            return

        point = self.window.view.intersect_ray_board(ray)
        if self.window.view.get_intersection_center(point):
            if not self.window.layout.is_valid(self.window.state):
                return
            self.__class__.first_time = False
            clock.unschedule(self.do_hint2)
            def delay(dt):
                players = self.window.layout.get_players(self.window, 
                                                         self.window.state)
                self.window.pop_control()
                self.window.menu_state = self.window.state
                self.window.players = players
                self.window.do_game()
            clock.schedule_once(delay, .5)
            self.inner_dz = -2

    def update_inner_disc(self):
        if self.window.layout.is_valid(self.window.state):
            self.inner_dz = 2
            if not self.first_time:
                clock.unschedule(self.do_hint2)
                clock.schedule_once(self.do_hint2, 8)
        else:
            self.inner_dz = -2

    def after_apply(self):
        self.update_inner_disc()
        clock.unschedule(self.do_hint1)

        # jump to default computer setting.
        if self.window.state[2] != susan.CELL_EMPTY:
            self.window.apply(susan.MovePiece(2, 6))
        if self.window.state[4] != susan.CELL_EMPTY:
            self.window.apply(susan.MovePiece(4, 14))

class Player(Controller):
    def __init__(self, win, player):
        self.window = win
        self.player = player

    def begin_game(self):
        pass

    def begin_turn(self):
        pass

    def after_apply(self):
        self.window.next_player()

class HumanPlayer(Player):
    def on_mouse_press(self, x, y, button, modifiers):
        ray = self.window.select_ray(x, y)
        cell, pos = self.window.view.get_intersection_piece(ray)
        if cell is not None:
            if self.window.state[cell] != self.player:
                return
            self.window.view.drag_cell = cell
            self.window.push_control(DragPiece(self.window, cell, pos))
            self.window.dispatch_event('on_mouse_drag', x, y, 0, 0, 0, 0)
            return

        point = self.window.view.intersect_ray_board(ray)
        ring = self.window.view.get_intersection_ring(point)
        if ring is not None:
            angle = self.window.view.get_intersection_angle(point)
            self.window.push_control(DragRing(self.window, ring, angle)) 

        if self.window.view.get_intersection_center(point):
            self.window.pop_control()
            self.window.do_menu()

class ComputerThread(threading.Thread):
    def __init__(self, layout, state, player, depth):
        super(ComputerThread, self).__init__()

        self.layout = layout
        self.state = state
        self.player = player
        self.depth = depth
        self.lock = threading.Lock()
        self.turn = None

    def run(self):
        turn = self.layout.best_turn(self.state, self.player, self.depth)
        self.lock.acquire()
        self.turn = turn
        self.lock.release()

    def get_turn(self):
        self.lock.acquire()
        turn = self.turn
        self.lock.release()
        return turn

class ComputerPlayer(Player):
    def __init__(self, win, player, depth):
        super(ComputerPlayer, self).__init__(win, player)
        self.depth = depth
        self.thread = None

    def update(self, dt):
        if self.thread:
            turn = self.thread.get_turn()
            if turn:
                clock.set_fps_limit(None)
                self.thread = None
                self.window.apply(turn)

    def begin_turn(self):
        super(ComputerPlayer, self).begin_turn()

        if self.depth >= 3:
            self.window.indicator_thinking = True

        def delay(dt):
            clock.set_fps_limit(30)
            self.thread = ComputerThread(self.window.layout, self.window.state,
                self.player, self.depth)
            self.thread.start()
        clock.schedule_once(delay, 1.)

    def on_mouse_press(self, x, y, button, modifiers):
        ray = self.window.select_ray(x, y)
        point = self.window.view.intersect_ray_board(ray)
        if self.window.view.get_intersection_center(point):
            clock.set_fps_limit(None)
            self.window.pop_control()
            self.window.do_menu()

class TutorComputerPlayer(ComputerPlayer):
    def __init__(self, win, player):
        super(TutorComputerPlayer, self).__init__(win, player, 1)

    def begin_game(self):
        self.window.do_hint(getres('hint1.png'), 
                            getres('hint2.png'), 
                            getres('hint3.png'))

class AnimateRing(Controller):
    def __init__(self, win, ring, to_angle):
        self.window = win
        self.ring = ring
        self.from_angle = win.view.ring_angles[ring]
        self.to_angle = to_angle
        self.period = max(abs(self.to_angle - self.from_angle) / 180., .001)
        self.t = 0

    def update(self, dt):
        self.t += dt / self.period
        self.window.view.ring_angles[self.ring] = \
            interp(self.from_angle, self.to_angle, self.t)
        if self.t >= 1.:
            self.window.pop_control()
            self.window.view.save_ring_angles()
            self.window.view.ring_angles[self.ring] = 0.
            self.finished()

    def finished(self):
        pass

class AnimateTwistRingTurn(AnimateRing):
    def __init__(self, win, turn):
        self.turn = turn
        to_angle = win.view.get_angle_for_twist(turn.ring, turn.twist)
        super(AnimateTwistRingTurn, self).__init__(win, turn.ring, to_angle)

    def finished(self):
        self.window.apply_immediate(self.turn)

class AnimatePiece(Controller):
    def __init__(self, win, from_pos, to_cell):
        self.window = win
        to_pos = win.view.get_cell_pos(to_cell)
        self.from_pos = from_pos
        self.move_vec = to_pos - self.from_pos
        self.period = max(abs(self.move_vec) / 8, 0.001)
        self.t = 0

    def update(self, dt):
        self.t += dt / self.period
        self.window.view.drag_piece_pos = self.from_pos + self.t * self.move_vec
        if self.t >= 1:
            self.window.pop_control()
            self.window.view.drag_cell = None
            self.window.view.drag_piece_pos = None
            self.finished()

    def finished(self):
        pass

class AnimateMovePieceTurn(AnimatePiece):
    def __init__(self, win, turn):
        if win.view.drag_piece_pos is not None:
            from_pos = win.view.drag_piece_pos
        else:
            from_pos = win.view.get_cell_pos(turn.cellfrom)
            win.view.drag_cell = turn.cellfrom
            win.view.drag_piece_pos = from_pos
        self.turn = turn
        super(AnimateMovePieceTurn, self).__init__(win, from_pos, turn.cellto)

    def finished(self):
        self.window.apply_immediate(self.turn)

class DragPiece(Controller):
    def __init__(self, win, cellfrom, ref_point):
        self.window = win
        self.cellfrom = cellfrom
        self.ref_point = ref_point

    def on_mouse_press(self, x, y, button, modifiers):
        return True

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        ray = self.window.select_ray(x, y)
        point = self.window.view.intersect_ray_board(ray)
        self.window.view.drag_piece_pos = point - self.ref_point

        cellto = self.window.view.get_intersection_cell(point)
        self.window.view.drag_piece_valid = cellto is not None and (
            self.window.layout.can_move(
                self.window.state, self.cellfrom, cellto) or
            cellto == self.cellfrom)

    def on_mouse_release(self, x, y, button, modifiers):
        ray = self.window.select_ray(x, y)
        point = self.window.view.intersect_ray_board(ray)
        cellto = self.window.view.get_intersection_cell(point)
        self.window.pop_control()
        if cellto is not None and self.window.layout.can_move(
                self.window.state, self.cellfrom, cellto):
            self.window.apply(susan.MovePiece(self.cellfrom, cellto))
        else:
            self.window.push_control(AnimatePiece(
                self.window, self.window.view.drag_piece_pos, self.cellfrom))
        
class DragRing(Controller):
    def __init__(self, win, ring, ref_angle):
        self.window = win
        self.ring = ring
        self.ref_angle = ref_angle - self.window.view.ring_angles[self.ring]

    def get_angle(self, x, y):
        ray = self.window.select_ray(x, y)
        point = self.window.view.intersect_ray_board(ray)
        angle = self.window.view.get_intersection_angle(point) - self.ref_angle
        if angle < 180.:
            angle += 360.
        if angle > 180.:
            angle -= 360.
        min_angle, max_angle = self.window.view.get_ring_angle_limits(self.ring)
        return min(max(angle, min_angle), max_angle)

    def on_mouse_press(self, x, y, button, modifiers):
        return True

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        angle = self.get_angle(x, y)
        self.window.view.ring_angles[self.ring] = angle
    
    def on_mouse_release(self, x, y, button, modifiers):
        angle = self.get_angle(x, y)
        twist = self.window.view.get_twist_for_angle(self.ring, angle)
        self.window.pop_control()
        if twist:
            self.window.apply(susan.TwistRing(self.ring, twist))
        else:
            self.window.push_control(AnimateRing(self.window, self.ring, 0.))

class GameOverDialog(BlockInputControl):
    def __init__(self, win):
        self.window = win
        scores = win.layout.get_player_scores(self.window.state)
        if scores[0] == scores[1]:
            self.is_draw = True
            self.win_player = 0
        else:
            self.win_player = scores.index(win.layout.goal_score)
            self.is_draw = False
        self.t = 0
        self.x = self.window.indicator_x
        self.y = self.window.indicator_y
        self.scale = self.window.indicator_scale

    def update(self, dt):
        self.window.indicator_ry += dt * 180 
        if self.window.indicator_ry > 90:
            self.window.indicator_ry -= 180.
            self.window.indicator_player = self.win_player
            if self.is_draw:
                self.win_player = 1 - self.win_player

        self.t += dt / 3
        self.t = min(1, self.t)
        self.window.indicator_x = interp(self.x, self.window.width/2, self.t)
        self.window.indicator_y = interp(self.y, self.window.height/2, self.t)
        self.window.indicator_scale = \
            interp(self.scale, self.window.width/4, self.t)
        return True

    def on_mouse_press(self, x, y, button, modifiers):
        if self.t < 1:
            return
        self.window.draw_indicator = False
        self.window.reset_indicator()
        self.window.pop_control()
        self.window.do_menu()

def main():
    configs = [
        Config(double_buffer=True, depth_size=24,
               sample_buffers=1, samples=4),
        Config(double_buffer=True, depth_size=24,
               sample_buffers=1, samples=2),
        Config(double_buffer=True, depth_size=24),
        Config(double_buffer=True, depth_size=16),
    ]
    for config in configs:
        try:
            w = GameWindow(config=config, resizable=True, caption='Lazy Susan')
            break
        except window.NoSuchConfigException:
            pass
    try:
        glEnable(GL_MULTISAMPLE_ARB)
    except:
        pass

    glEnable(GL_DEPTH_TEST)
    w.begin_intro()
    while not w.has_exit:
        w.dispatch_events()
        dt = clock.tick()
        if dt > .1:
            dt = 0.01
        w.update(dt)
        w.draw()

if __name__ == '__main__':
    main()
